#ifndef VNET_YARP_TABLE_HH_
#define VNET_YARP_TABLE_HH_

#include <vector>
#include <map>
#include <utility>

#include <boost/shared_ptr.hpp>
#include <boost/thread/mutex.hpp>

#include <vnet_filter.hh>

namespace vnet {
	namespace yarp {


		class table {
			public:
				typedef std::pair<std::string, std::string> robot_pair;
				typedef boost::shared_ptr<filter_base> filter_ptr;
				typedef std::vector<filter_ptr> filter_array;
				typedef std::map<robot_pair, filter_array> filter_map;

			private:
				boost::mutex mtx;
				filter_map m;
				filter_array default_array;

				void add(const robot_pair& p, filter_base *f)
				{
					filter_map::iterator it = m.find(p);
					if (it == m.end()) {
						filter_array arr;
						arr.push_back(filter_ptr(f));
						m[p] = arr;
					} else {
						it->second.push_back(filter_ptr(f));
					}
				}

				void remove(const robot_pair& p, int index)
				{
					filter_map::iterator it = m.find(p);
					if (it == m.end()) {
						return;
					} else {
						if (index >= it->second.size())
							return;
						else {
							it->second.erase(it->second.begin() + index);
							if (it->second.size() == 0)
								m.erase(it);
						}
					}
				}

				public:
					table(filter_base* default_filter) {
						default_array.push_back(filter_ptr(default_filter));
					}

					void add(const std::string& first, const std::string& snd,
							 filter_base * f, bool bi = false) 
					{
						add(std::make_pair(first, snd), f);
						if (bi) 
							add(std::make_pair(snd, first), f);
					}

					const filter_array& 
					get(const std::string& first, const std::string& snd) const
					{
						return get(std::make_pair(first, snd));
					}

					const filter_array&
					get(const robot_pair& p) const
					{
						filter_map::const_iterator it = m.find(p);
						if (it == m.end())
							return default_array;
						else
							return it->second;
					}

					void clear(const std::string& first, const std::string& snd, 
							   bool bi = false)
					{
						m.erase(std::make_pair(first, snd));
						if (bi)
							m.erase(std::make_pair(snd, first));
					}

					void remove(const std::string& first, const std::string& snd,
								int index, bool bi = false)
					{
						remove(std::make_pair(first, snd), index);
						if (bi)
							remove(std::make_pair(snd, first), index);
					}

					std::string repr() const
					{
						std::ostringstream oss;
						for (filter_map::const_iterator it = m.begin();
							 it != m.end(); 
							 ++it)
						{
							oss << "<" << it->first.first << ", " << it->first.second;
							oss << "> :\n";
							for (size_t i = 0; i < it->second.size(); ++i) 
								oss << "\t" << i << ": " << it->second[i]->repr() << "\n";
						}
						
						return oss.str();
					}

					void lock() {
						mtx.lock();
					}

					void unlock() {
						mtx.unlock();
					}

					bool try_lock() {
						return mtx.try_lock();
					}
						
			};
		}
	}


#endif /* VNET_YARP_TABLE_HH_ */
